@@ -1,7 +1,9 @@
 use json::JsonValue;
-use log::{debug, warn};
+use log::debug;
+use percent_encoding::percent_decode_str;
+use url::Url;
 
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::net::{Ipv4Addr, Ipv6Addr, SocketAddr, ToSocketAddrs};
 
 #[derive(Debug, Clone)]
@@ -31,11 +33,47 @@ pub type IpMask = (Ipv6Addr, u32);
 
 #[derive(Debug, Clone)]
 pub struct ServerTlsConfig {
+    pub sni_hostnames: HashSet<TlsOption>,
+    // the alpn protocols to show in the serverhello response
+    pub alpn_protocols: HashSet<TlsOption>,
+
     pub cert_path: String,
     pub key_path: String,
     pub optional: bool,
 }
 
+#[derive(Debug, PartialEq, Eq, Hash, Clone)]
+pub enum TlsOption {
+    None,
+    Any,
+    Specified(String),
+}
+
+impl From<&str> for TlsOption {
+    fn from(s: &str) -> Self {
+        match s {
+            "any" => TlsOption::Any,
+            "none" => TlsOption::None,
+            hostname => TlsOption::Specified(hostname.to_string()),
+        }
+    }
+}
+
+impl TlsOption {
+    pub fn is_specified(&self) -> bool {
+        match self {
+            TlsOption::Specified(_) => true,
+            _ => false,
+        }
+    }
+    pub fn unwrap_specified(&self) -> &str {
+        match self {
+            TlsOption::Specified(s) => s.as_str(),
+            _ => panic!("Not a hostname"),
+        }
+    }
+}
+
 #[derive(Debug, Clone)]
 pub struct ClientTlsConfig {
     pub verify: bool,
@@ -43,10 +81,9 @@ pub struct ClientTlsConfig {
 
 #[derive(Debug, Clone)]
 pub struct TcpTargetConfig {
+    pub allowed_ips: Vec<IpMask>,
     pub target_addresses: Vec<TcpTargetAddress>,
-    pub allowlist: Vec<IpMask>,
     pub server_tls_config: Option<ServerTlsConfig>,
-    pub early_connect: bool,
     pub tcp_nodelay: bool,
 }
 
@@ -69,24 +106,49 @@ impl std::fmt::Display for UdpTargetAddress {
 #[derive(Debug, Clone)]
 pub struct UdpTargetConfig {
     pub target_addresses: Vec<UdpTargetAddress>,
-    pub allowlist: Vec<IpMask>,
+    pub allowed_ips: Vec<IpMask>,
     pub association_timeout_secs: Option<u32>,
 }
 
-pub fn load_configs(config_paths: Vec<String>, config_urls: Vec<String>) -> Vec<ServerConfig> {
-    let mut config_objects: Vec<JsonValue> = config_paths
-        .iter()
-        .map(|config_path| {
-            let config_str =
-                std::fs::read_to_string(config_path).expect("Failed to read from config file");
-            let config_str = config_str
-                .split('\n')
-                .filter(|line| return !line.trim_start().starts_with("//"))
-                .collect::<Vec<&str>>()
-                .join("\n");
-            json::parse(&config_str).expect("Invalid JSON in config")
-        })
-        .collect();
+pub async fn load_configs(
+    config_paths: Vec<String>,
+    config_urls: Vec<String>,
+    is_initial: bool,
+) -> Vec<ServerConfig> {
+    let mut config_objects = vec![];
+    for config_path in config_paths {
+        let config_str = match tokio::fs::read_to_string(&config_path).await {
+            Ok(s) => s,
+            Err(e) => {
+                println!("Failed to read {}, skipping: {}", config_path, e);
+                if is_initial {
+                    panic!("Initial load failed.");
+                }
+                continue;
+            }
+        };
+        let config_str = config_str
+            .split('\n')
+            .filter(|line| return !line.trim_start().starts_with("//"))
+            .collect::<Vec<&str>>()
+            .join("\n");
+        let config_object = match json::parse(&config_str) {
+            Ok(o) => o,
+            Err(e) => {
+                println!("Failed to read {}, invalid JSON: {}", config_path, e);
+                if is_initial {
+                    panic!("Initial load failed.");
+                }
+                continue;
+            }
+        };
+        config_objects.push(config_object);
+    }
+
+    for config_url in config_urls {
+        let config_object = convert_url_to_obj(&config_url).unwrap();
+        config_objects.push(config_object);
+    }
 
     let mut ip_groups = HashMap::new();
     for config_object in config_objects.iter_mut() {
@@ -101,10 +163,6 @@ pub fn load_configs(config_paths: Vec<String>, config_urls: Vec<String>) -> Vec<
         all_configs.extend(configs.into_iter())
     }
 
-    for config_url in config_urls {
-        all_configs.push(load_config_from_url(config_url, &ip_groups));
-    }
-
     all_configs
 }
 
@@ -177,87 +235,6 @@ fn convert_ip_mask(mask_str: &str) -> IpMask {
     }
 }
 
-fn load_config_from_url(url: String, ip_groups: &HashMap<String, Vec<IpMask>>) -> ServerConfig {
-    fn split(mut input: String, separator: &str) -> (String, String) {
-        match input.find(separator) {
-            Some(index) => {
-                let suffix = input.split_off(index + separator.len());
-                input.truncate(input.len() - separator.len());
-                (input, suffix)
-            }
-            None => (input, String::new()),
-        }
-    }
-
-    fn trim(input: &mut String, c: u8) {
-        let b = input.as_bytes();
-        let mut i = b.len() - 1;
-        while b[i] == c {
-            i -= 1;
-        }
-        input.truncate(i + 1);
-    }
-
-    fn b(input: String) -> bool {
-        input != "false" && input != "0"
-    }
-
-    let mut obj = JsonValue::new_object();
-
-    let (protocol, remaining) = split(url, "://");
-
-    obj.insert("protocol", protocol).unwrap();
-
-    let (mut address, query) = split(remaining, "?");
-    trim(&mut address, b'/');
-
-    obj.insert("bindAddress", address.clone()).unwrap();
-
-    // only a single target is supported with urls.
-    let mut target = JsonValue::new_object();
-
-    for param in query.split("&") {
-        let (key, value) = split(param.to_string(), "=");
-        if key == "iptables" {
-            obj.insert("iptables", b(value)).unwrap();
-        } else if key == "targetAddress" || key == "target" || key == "to" {
-            target.insert("address", value).unwrap();
-        } else if key == "tcp_nodelay" || key == "nodelay" || key == "tcpNodelay" {
-            target.insert("tcp_nodelay", b(value)).unwrap();
-        } else if key == "early_connect" || key == "earlyConnect" {
-            target.insert("early_connect", b(value)).unwrap()
-        } else if key == "allowlist" || key == "allow" || key == "allowed" {
-            target
-                .insert(
-                    "allowlist",
-                    value
-                        .split(",")
-                        .map(|item| item.to_string())
-                        .collect::<Vec<String>>(),
-                )
-                .unwrap();
-        } else {
-            panic!("Unknown query parameter: {}", key);
-        }
-    }
-
-    if !target.has_key("address") {
-        panic!("Config URL for {} is missing target address.", address);
-    }
-
-    if !target.has_key("allowlist") {
-        warn!(
-            "Config URL for {} missing allowlist, allowing all connections.",
-            address
-        );
-        target.insert("allowlist", vec!["0.0.0.0/0"]).unwrap();
-    }
-
-    obj.insert("target", target).unwrap();
-
-    parse_server_object(obj, ip_groups)
-}
-
 fn load_config(mut obj: JsonValue, ip_groups: &HashMap<String, Vec<IpMask>>) -> Vec<ServerConfig> {
     let server_objs = match obj {
         JsonValue::Array(v) => v,
@@ -294,37 +271,35 @@ fn parse_server_object(
         .next()
         .expect("Unable to resolve bind address");
 
-    let use_iptables = obj["iptables"].as_bool().unwrap_or(false);
+    let use_iptables = is_true_value(&obj["iptables"], false);
+
+    let target_value = take_existing(&mut obj, &["target", "targets", "to"]);
+
+    let target_objs = match target_value {
+        JsonValue::String(s) => {
+            let mut json_obj = JsonValue::new_object();
+            json_obj["address"] = s.into();
+            vec![json_obj]
+        }
+        JsonValue::Object(_) => vec![target_value],
+        JsonValue::Array(v) => v,
+        _ => panic!("Invalid targets"),
+    };
 
     let target_configs = match obj["protocol"].as_str().unwrap_or("tcp") {
         "tcp" => {
-            let target_configs = if obj.has_key("target") {
-                vec![parse_tcp_target_object(obj["target"].take(), ip_groups)]
-            } else {
-                let target_objs = match obj["targets"].take() {
-                    JsonValue::Array(v) => v,
-                    _ => panic!("Invalid targets"),
-                };
-                target_objs
-                    .into_iter()
-                    .map(|target_obj| parse_tcp_target_object(target_obj, ip_groups))
-                    .collect()
-            };
+            let target_configs = target_objs
+                .into_iter()
+                .map(|target_obj| parse_tcp_target_object(target_obj, ip_groups))
+                .collect();
+
             TargetConfigs::Tcp(target_configs)
         }
         "udp" => {
-            let target_configs = if obj.has_key("target") {
-                vec![parse_udp_target_object(obj["target"].take(), ip_groups)]
-            } else {
-                let target_objs = match obj["targets"].take() {
-                    JsonValue::Array(v) => v,
-                    _ => panic!("Invalid targets"),
-                };
-                target_objs
-                    .into_iter()
-                    .map(|target_obj| parse_udp_target_object(target_obj, ip_groups))
-                    .collect()
-            };
+            let target_configs = target_objs
+                .into_iter()
+                .map(|target_obj| parse_udp_target_object(target_obj, ip_groups))
+                .collect();
             TargetConfigs::Udp(target_configs)
         }
         unknown_protocol => {
@@ -362,26 +337,8 @@ fn parse_tcp_target_object(
         panic!("No target addresses specified.");
     }
 
-    let allowlist = match obj["allowlist"].take() {
-        JsonValue::String(s) => lookup_ip_mask(&s, ip_groups),
-        JsonValue::Short(s) => lookup_ip_mask(s.as_str(), ip_groups),
-        JsonValue::Array(v) => v
-            .into_iter()
-            .map(|v| lookup_ip_mask(v.as_str().expect("Invalid allowlist entry"), ip_groups))
-            .collect::<Vec<Vec<IpMask>>>()
-            .concat(),
-        invalid => panic!("Invalid allowlist value: {}", invalid),
-    };
-
-    let early_connect = match obj["early_connect"].take() {
-        JsonValue::Boolean(b) => b,
-        JsonValue::Null => false,
-        invalid => panic!("Invalid early_connect value: {}", invalid),
-    };
-
-    if early_connect {
-        warn!("Enabling early connect, this can cause excessive target connections on unauthorized connections.");
-    }
+    // support allowlist key for backwards compatibility
+    let allowed_ips = parse_allowed_ips(&mut obj, ip_groups);
 
     let tcp_nodelay = match obj["tcp_nodelay"].take() {
         JsonValue::Boolean(b) => b,
@@ -390,14 +347,34 @@ fn parse_tcp_target_object(
     };
 
     TcpTargetConfig {
+        allowed_ips,
         server_tls_config,
         target_addresses,
-        allowlist,
-        early_connect,
         tcp_nodelay,
     }
 }
 
+fn parse_allowed_ips(obj: &mut JsonValue, ip_groups: &HashMap<String, Vec<IpMask>>) -> Vec<IpMask> {
+    let allowed_ips_obj = take_existing(obj, &["allowed_ips", "allowlist"]);
+
+    let allowed_ip_strs = match allowed_ips_obj {
+        JsonValue::String(s) => vec![s],
+        JsonValue::Short(s) => vec![s.as_str().to_string()],
+        JsonValue::Array(v) => v
+            .into_iter()
+            .map(|v| v.as_str().expect("Invalid allowed_ips entry").to_string())
+            .collect::<Vec<_>>(),
+        JsonValue::Null => vec!["all".to_string()],
+        invalid => panic!("Invalid allowed_ips value: {}", invalid),
+    };
+
+    allowed_ip_strs
+        .into_iter()
+        .map(|s| lookup_ip_mask(&s, ip_groups))
+        .collect::<Vec<Vec<IpMask>>>()
+        .concat()
+}
+
 fn lookup_ip_mask(s: &str, ip_groups: &HashMap<String, Vec<IpMask>>) -> Vec<IpMask> {
     if s.starts_with("@") {
         match ip_groups.get(&s[1..]) {
@@ -414,11 +391,53 @@ fn lookup_ip_mask(s: &str, ip_groups: &HashMap<String, Vec<IpMask>>) -> Vec<IpMa
 fn parse_server_tls_object(obj: JsonValue) -> Option<ServerTlsConfig> {
     match obj {
         JsonValue::Null => None,
-        JsonValue::Object(mut o) => Some(ServerTlsConfig {
-            cert_path: o["cert"].take_string().expect("No cert path"),
-            key_path: o["key"].take_string().expect("No key path"),
-            optional: o["optional"].as_bool().unwrap_or(false),
-        }),
+        JsonValue::Object(mut o) => {
+            let sni_hostnames = match o["sni_hostnames"].take() {
+                JsonValue::String(s) => HashSet::from([s.as_str().into()]),
+                JsonValue::Short(s) => HashSet::from([s.as_str().into()]),
+                JsonValue::Array(v) => {
+                    let mut hostnames = v
+                        .into_iter()
+                        .map(|v| v.as_str().expect("Invalid sni_hostnames entry").into())
+                        .collect::<HashSet<_>>();
+                    if hostnames.is_empty() {
+                        hostnames.extend([TlsOption::None, TlsOption::Any]);
+                    }
+                    hostnames
+                }
+                JsonValue::Null => HashSet::from([TlsOption::None, TlsOption::Any]),
+                invalid => panic!("Invalid sni_hostnames value: {}", invalid),
+            };
+
+            let cert_path = o["cert"].take_string().expect("No cert path");
+            let key_path = o["key"].take_string().expect("No key path");
+            let optional = is_true_value(&o["optional"], false);
+
+            let alpn_protocols = match o["alpn_protocols"].take() {
+                JsonValue::String(s) => HashSet::from([s.as_str().into()]),
+                JsonValue::Short(s) => HashSet::from([s.as_str().into()]),
+                JsonValue::Array(v) => {
+                    let mut hostnames = v
+                        .into_iter()
+                        .map(|v| v.as_str().expect("Invalid alpn_protocols entry").into())
+                        .collect::<HashSet<_>>();
+                    if hostnames.is_empty() {
+                        hostnames.extend([TlsOption::None, TlsOption::Any]);
+                    }
+                    hostnames
+                }
+                JsonValue::Null => HashSet::from([TlsOption::None, TlsOption::Any]),
+                invalid => panic!("Invalid alpn_protocols value: {}", invalid),
+            };
+
+            Some(ServerTlsConfig {
+                sni_hostnames,
+                alpn_protocols,
+                cert_path,
+                key_path,
+                optional,
+            })
+        }
         _ => {
             panic!("Unknown server TLS config");
         }
@@ -431,12 +450,12 @@ fn parse_client_tls_object(obj: JsonValue) -> Option<ClientTlsConfig> {
         JsonValue::Object(mut o) => {
             let verify = o
                 .remove("verify")
-                .map(|v| v.as_bool().unwrap_or(true))
+                .map(|v| is_true_value(&v, true))
                 .unwrap_or(true);
             Some(ClientTlsConfig { verify })
         }
         _ => {
-            if obj.as_bool().unwrap_or(false) {
+            if is_true_value(&obj, false) {
                 Some(ClientTlsConfig { verify: true })
             } else {
                 None
@@ -517,22 +536,13 @@ fn parse_udp_target_object(
         panic!("No target addresses specified.");
     }
 
-    let allowlist = match obj["allowlist"].take() {
-        JsonValue::String(s) => lookup_ip_mask(&s, ip_groups),
-        JsonValue::Short(s) => lookup_ip_mask(s.as_str(), ip_groups),
-        JsonValue::Array(v) => v
-            .into_iter()
-            .map(|v| lookup_ip_mask(v.as_str().expect("Invalid allowlist entry"), ip_groups))
-            .collect::<Vec<Vec<IpMask>>>()
-            .concat(),
-        invalid => panic!("Invalid allowlist value: {}", invalid),
-    };
+    let allowed_ips = parse_allowed_ips(&mut obj, ip_groups);
 
     let association_timeout_secs = obj["association_timeout_secs"].as_u32();
 
     UdpTargetConfig {
         target_addresses,
-        allowlist,
+        allowed_ips,
         association_timeout_secs,
     }
 }
@@ -562,3 +572,83 @@ fn parse_udp_target_address(mut obj: JsonValue) -> UdpTargetAddress {
         panic!("Invalid target address: {}", obj);
     }
 }
+
+fn is_true_value(value: &JsonValue, default_value: bool) -> bool {
+    if value.is_string() {
+        let value_str = value.as_str().unwrap();
+        return value_str == "true" || value_str == "1" || value_str == "yes";
+    } else if value.is_number() {
+        let value = value.as_i32().unwrap();
+        return value == 1;
+    }
+    value.as_bool().unwrap_or(default_value)
+}
+
+fn convert_url_to_obj(url_str: &str) -> std::result::Result<JsonValue, String> {
+    let url = Url::parse(url_str).map_err(|e| format!("Failed to parse url: {}", e))?;
+
+    let mut json_obj = JsonValue::new_object();
+
+    let protocol = url.scheme();
+
+    json_obj.insert("protocol", protocol.to_string()).unwrap();
+
+    let host_str = match url.host_str() {
+        Some(s) => s.to_string(),
+        None => {
+            return Err(format!("URL missing host: {}", url_str));
+        }
+    };
+
+    let address = match url.port() {
+        Some(port) => format!("{}:{}", host_str, port),
+        None => host_str,
+    };
+
+    json_obj.insert("bindAddress", address).unwrap();
+
+    for (query_key, query_value) in url.query_pairs().into_owned() {
+        let new_value = JsonValue::String(
+            percent_decode_str(&query_value)
+                .decode_utf8()
+                .unwrap()
+                .into_owned(),
+        );
+
+        let mut key_parts = query_key.split('.').collect::<Vec<_>>();
+        let final_part = key_parts.pop().unwrap();
+
+        let mut current_obj = &mut json_obj;
+        for key_part in key_parts.into_iter() {
+            if !current_obj.has_key(&key_part) {
+                current_obj[key_part] = JsonValue::new_object();
+            }
+            current_obj = &mut current_obj[key_part];
+        }
+
+        if current_obj.has_key(final_part) {
+            let existing_value = &mut current_obj[final_part];
+            if existing_value.is_array() {
+                existing_value.push(new_value).unwrap();
+            } else {
+                let existing_value = current_obj.remove(final_part);
+                current_obj
+                    .insert(final_part, vec![existing_value, new_value])
+                    .unwrap();
+            }
+        } else {
+            current_obj.insert(final_part, new_value).unwrap();
+        }
+    }
+
+    Ok(json_obj)
+}
+
+fn take_existing(obj: &mut JsonValue, keys: &[&str]) -> JsonValue {
+    for key in keys {
+        if obj.has_key(*key) {
+            return obj[*key].take();
+        }
+    }
+    JsonValue::Null
+}
